/*
 * LibNC
 * 
 * Copyright (c) 2018-2019 Fabrice Bellard
 *
 */
#ifndef LIBNC_H
#define LIBNC_H

#include <inttypes.h>
#include "cutils.h"

/* basic operations */

void *nc_malloc(size_t size);
void *nc_mallocz(size_t size);
void nc_free(void *ptr);

typedef struct {
    uint32_t seed;
    /* used by Gaussian generator */
    int idx;
    float y1;
} RNDState;

typedef struct {
    uint16_t u16;
} nc_float16_t;

void rnd_init(RNDState *s, uint32_t seed);
uint32_t rnd_unif_u32(RNDState *s);
float rnd_unif(RNDState *s);
void rnd_unif_vec(float *tab, size_t n, float mu, float range,
                  RNDState *s);
void rnd_unif_mat(float *tab, size_t stride, size_t h, size_t w,
                  float mu, float sigma, RNDState *s);

float vec_sum(const float *tab, size_t n);
void mat_zero(float *tab, size_t stride, size_t h, size_t w);
size_t get_mat_stride(size_t n);
struct NCThreadPool;
void mat_mul(struct NCThreadPool *ts, float *tab_c, size_t c_stride,
             const float *tab_a, size_t a_stride,
             const float *tab_b, size_t b_stride,
             BOOL a_trans, BOOL b_trans, int m, int n, int k);
void mat_mul_f16(float *tab_c, size_t c_stride,
                 const nc_float16_t *tab_a, size_t a_stride,
                 const float *tab_b, size_t b_stride,
                 size_t m, size_t n, size_t k);
void convert_f32_to_f16(nc_float16_t *dst, const float *src, size_t n);

/* profiling */

typedef enum {
    PROF_EVAL,
    PROF_GRAD,
    PROF_SGD,
    PROF_UPDATE,
    PROF_WRITE_SYM,
    PROF_TOTAL,
    PROF_COUNT,
} ProfEnum;

#ifdef PROFILE

extern int64_t prof_cycles[PROF_COUNT];
extern int64_t prof_samples[PROF_COUNT];
extern int64_t prof_ops[PROF_COUNT];

static inline void prof_start(int idx)
{
    prof_cycles[idx] -= get_cycles();
}

static inline void prof_end(int idx)
{
    prof_cycles[idx] += get_cycles();
    prof_samples[idx]++;
}

static inline void prof_end_ops(int idx, int n_ops)
{
    prof_cycles[idx] += get_cycles();
    prof_ops[idx] += n_ops;
    prof_samples[idx]++;
}

#else

static inline void prof_start(int idx)
{
}

static inline void prof_end(int idx)
{
}

static inline void prof_end_ops(int idx, int n_ops)
{
}

#endif

void nc_prof_dump(void);

/* Automatic Differentiation Engine */

typedef struct NCModel NCModel;
typedef struct NCFunction NCFunction;
typedef struct NCSignal NCSignal;
typedef struct NCSGDOptState NCSGDOptState;

typedef enum {
    NC_TYPE_F32,
    NC_TYPE_I32,
    NC_TYPE_F16,
    NC_TYPE_COUNT,
} NCTypeEnum;

typedef struct NCSignalData {
    NCTypeEnum item_type;
    size_t item_size;
    void *data;
    size_t stride; /* in elements */
    size_t n_strides; /* prod(j = 1 ... n_dims, dims[j]); */
    int n_dims;
    const size_t *dims; /* n_dims length */
    const size_t *strides; /* n_dims length, strides in bytes */
} NCSignalData;

typedef struct {
    /* if true, 'grad_func' is compiled. It evaluates the gradient of
       the function and invoke the gradient optimizer to update each
       variable.
     */
    BOOL grad;
    
    /* if true, partial_eval_func[] is compiled. partial_eval_func[i]
       can be called to evaluate output[0 ... i] on axis
       'output_axis' */
    BOOL partial_eval;
    NCSignal *output; /* necessary if partial_eval = true */
    int output_axis; /* necessary if partial_eval = true */
} NCCompileParams;

typedef struct {
    NCFunction *eval_func; /* evaluation function */
    NCFunction *grad_func; /* gradient function */
    NCFunction **partial_eval_func; /* partial evaluation functions */
} NCCompileResult;

NCModel *nc_model_init(int nb_threads);
void nc_model_end(NCModel *m);

void nc_start_func(NCModel *m, const char *name);
/* compile the function whose definition was started by nc_start_func() */
void nc_compile(NCModel *m, NCCompileResult *cr, const NCCompileParams *p);
void nc_free_func(NCFunction *f);

/* evaluate the compiled function 'f' */
void nc_eval(NCFunction *f);

/* finish the variable update with the computed gradients and the SGD
   optimizers */
void nc_update(NCModel *m);

NCSignal *nc_new_global(NCModel *m, NCTypeEnum type,
                        int n_dims, const size_t *dims);
NCSignal *nc_new_global_1d(NCModel *m, NCTypeEnum type, size_t len);
NCSignal *nc_new_global_2d(NCModel *m, NCTypeEnum type, size_t n0, size_t n1);
NCSignal *nc_new_global_3d(NCModel *m, NCTypeEnum type,
                           size_t n0, size_t n1, size_t n2);
NCSignal *nc_new_var(NCModel *m, NCTypeEnum type,
                     int n_dims, const size_t *dims);
NCSignal *nc_new_var_1d(NCModel *m, NCTypeEnum type, size_t len);
NCSignal *nc_new_var_2d(NCModel *m, NCTypeEnum type, size_t n0, size_t n1);
NCSignal *nc_new_var_3d(NCModel *m, NCTypeEnum type,
                        size_t n0, size_t n1, size_t n2);
NCSignal *nc_new_slice(NCSignal *x1, int axis, int start, int end);
NCSignal *nc_signal_set_global(NCSignal *x);
NCSignal *nc_signal_set_name_str(NCSignal *x, const char *name);
NCSignal *__attribute__((format(printf, 2, 3))) nc_signal_set_name(NCSignal *x, const char *fmt, ...);
void nc_free_var(NCSignal *x);
NCSignal *nc_inc_ref(NCSignal *x);
void nc_dec_ref(NCSignal *x);
NCSignal *nc_find_param(NCModel *m, const char *name);

NCTypeEnum nc_signal_get_item_type(NCSignal *x);
NCSignalData *nc_signal_get_data(NCSignalData *sd, NCSignal *x);
void nc_signal_set_zero(NCSignal *y);
void nc_signal_set_f32(NCSignal *y, float val);
void nc_signal_copy(NCSignal *dst, NCSignal *src);
void nc_signal_convert(NCSignal *dst, NCSignal *src);

void nc_dump_dims(const char *str, NCSignal *x);
void nc_dump_params(NCModel *m);
void nc_dump_globals(NCModel *m);
int nc_get_param_count(NCModel *m);
size_t nc_get_heap_size(NCModel *m);
void nc_save_coefs(NCModel *m, const char *filename);
void nc_load_coefs(NCModel *m, const char *filename);
void nc_save_state(NCModel *m, const char *filename);
void nc_load_state(NCModel *m, const char *filename);

/* operations */
NCSignal *nc_set_vec(NCModel *m, size_t n, float val);
NCSignal *nc_convert(NCSignal *x, NCTypeEnum new_type);
NCSignal *nc_reshape(NCSignal *x, int n_dims, const size_t *dims);
NCSignal *nc_reshape_1d(NCSignal *x, size_t n0);
NCSignal *nc_reshape_2d(NCSignal *x, size_t n0, size_t n1);
NCSignal *nc_reshape_3d(NCSignal *x, size_t n0, size_t n1, size_t n2);
NCSignal *nc_dup(NCSignal *x, size_t n);
NCSignal *nc_add(NCSignal *x1, NCSignal *x2);
NCSignal *nc_mul(NCSignal *x1, NCSignal *x2);
NCSignal *nc_mul_const(NCSignal *x, float c);
NCSignal *nc_max(NCSignal *x1, NCSignal *x2);
/* concatenation along axis 'axis' */
NCSignal *nc_concat(NCSignal **inputs, int n_inputs, int axis);
/* shortcut for axis = 0 */
NCSignal *nc_vconcat(NCSignal **inputs, int n_inputs);
/* shortcut for axis = 1 */
NCSignal *nc_hconcat(NCSignal **inputs, int n_inputs);
/* split along axis 'axis'. If tab_size = NULL, split equally. */
void nc_split(NCSignal **tab_y, NCSignal *x, int n_outputs,
              const size_t *tab_size, int axis);
/* shortcut for axis = 0 */
void nc_vsplit(NCSignal **tab_y, NCSignal *x, int n_outputs,
               const size_t *tab_size);
/* shortcut for axis = 1 */
void nc_hsplit(NCSignal **tab_y, NCSignal *x, int n_outputs,
               const size_t *tab_size);
/* split 'x' along the axis 0 into n equal parts and concatenate the
   result in a single array with one more dimension.  */
NCSignal *nc_vsplit_dim(NCSignal *x, int n);
/* inverse of vsplit_dim */
NCSignal *nc_vconcat_dim(NCSignal *x);

typedef enum {
    NC_PAD_ZERO,
    NC_PAD_DUP, /* duplicate element */
    /* trim types, dual to padding */
    NC_TRIM_NORMAL = NC_PAD_ZERO,
    NC_TRIM_SUM, /* add trimmed elements to the edge */
} NCPadEnum;

/* pad (len > 0) or trim (len < 0) the axis 0 of 'x' */
NCSignal *nc_pad(NCSignal *x, ssize_t left_len, NCPadEnum left_op,
                 ssize_t right_len, NCPadEnum right_op);
/* shortcut to nc_pad() */
NCSignal *nc_resize(NCSignal *x, size_t n);

NCSignal *nc_transpose(NCSignal *x);
NCSignal *nc_matmul(NCSignal *w, NCSignal *x);

typedef enum {
    /* no special hint for the partial evaluation */
    NC_MATMUL_HINT_NONE,
    /* assume that k-n+1+i zeros are in the lower part of x2
       column i = 0 ... n - 1 */
    NC_MATMUL_HINT1,
    /* assume that the lower triangular part of the result
       won't be used. The partial multiplication gives the
       right result even if k changes because only trailing
       zeros are suppressed. */
    NC_MATMUL_HINT2,
} NCMatMulHintEnum;

NCSignal *nc_matmul_hint(NCSignal *a, NCSignal *b, BOOL a_trans, BOOL b_trans,
                         NCMatMulHintEnum hint);
NCSignal *nc_matmul_stride(NCSignal *w, NCSignal *x);
/* return a matrix where each column is the column x[i] of matrix 'w' */
NCSignal *nc_get_col(NCSignal *w, NCSignal *x);
NCSignal *nc_soft_max(NCSignal *x);
NCSignal *nc_log_soft_max(NCSignal *x, NCSignal *eout);
NCSignal *nc_sigmoid(NCSignal *x);
NCSignal *nc_tanh(NCSignal *x);
NCSignal *nc_relu(NCSignal *x);
NCSignal *nc_layer_norm(NCSignal *x, float eps);
NCSignal *nc_rms_norm(NCSignal *x, float eps);
/* return cp * fg + min(1 - fg, ig) * in */
NCSignal *nc_lstm_clamped(NCSignal *cp, NCSignal *in,
                          NCSignal *fg, NCSignal *ig);
/* return cp * fg + (1 - fg) * in */
NCSignal *nc_interp(NCSignal *cp, NCSignal *in, NCSignal *fg);
NCSignal *nc_slt_mat_set(NCSignal *x, int pos, float c);
NCSignal *nc_rel_shift(NCSignal *x, int pos);

/* SGD optimizer */

typedef enum {
    SGD_OPT_BASIC,
    SGD_OPT_ADAM,
    SGD_OPT_TEST,
} SGDOptAlgoEnum;

typedef struct {
    SGDOptAlgoEnum algo;
    float learning_rate;
    union {
        struct {
            float beta1;
            float beta2;
            float eps;
        } adam;
    } u;
} SGDOptParams;

NCSGDOptState *nc_sgd_opt_init(NCModel *m, const SGDOptParams *p);
void nc_sgd_opt_end(NCSGDOptState *s);
void nc_sgd_opt_set_learning_rate(NCSGDOptState *s, float lr);

/* set the SGD optimizer 's' to all parameters of the model */
void nc_sgd_opt_set_all(NCModel *m, NCSGDOptState *s);

/* set the SGD optimizer 's' to the variable 'x'. Remove it if s = NULL */
void nc_sgd_opt_set(NCSignal *x, NCSGDOptState *s);

/* Return the gradient associated to the variable 'x'. Only works with
   'SGD_OPT_TEST' */
float *nc_signal_get_grad(NCSignal *x);

#endif /* LIBNC_H */
